@@ -44,6 +44,27 @@ MVM_PUBLIC void MVM_repr_pos_set_elems(MVMThreadContext *tc, MVMObject *obj, MVM
         OBJECT_BODY(obj), elems);
 }
 
+static void int_array_to_c_array(MVMThreadContext *tc, MVMObject *arr, MVMint64 *elems, MVMint64 **values) {
+    MVMint64 i;
+    *elems = MVM_repr_elems(tc, arr);
+    *values = *elems ? MVM_malloc(*elems * sizeof(MVMint64)) : NULL;
+    for (i = 0; i < *elems; i++)
+        (*values)[i] = MVM_repr_at_pos_i(tc, arr, i);
+}
+
+void MVM_repr_set_dimensions(MVMThreadContext *tc, MVMObject *obj, MVMObject *dims) {
+    if (IS_CONCRETE(obj)) {
+        MVMint64 num_dims, *c_dims;
+        int_array_to_c_array(tc, dims, &num_dims, &c_dims);
+        REPR(obj)->pos_funcs.set_dimensions(tc, STABLE(obj), obj,
+            OBJECT_BODY(obj), num_dims, c_dims);
+        MVM_free(c_dims);
+    }
+    else {
+        MVM_exception_throw_adhoc(tc, "Cannot set dimensions on a type object");
+    }
+}
+
 MVM_PUBLIC void MVM_repr_pos_splice(MVMThreadContext *tc, MVMObject *obj, MVMObject *replacement, MVMint64 offset, MVMint64 count) {
     REPR(obj)->pos_funcs.splice(tc, STABLE(obj), obj,
         OBJECT_BODY(obj), replacement,
@@ -51,8 +72,10 @@ MVM_PUBLIC void MVM_repr_pos_splice(MVMThreadContext *tc, MVMObject *obj, MVMObj
 }
 
 MVM_PUBLIC MVMint64 MVM_repr_exists_pos(MVMThreadContext *tc, MVMObject *obj, MVMint64 index) {
-    return REPR(obj)->pos_funcs.exists_pos(tc,
-        STABLE(obj), obj, OBJECT_BODY(obj), index);
+    MVMint64 elems = REPR(obj)->elems(tc, STABLE(obj), obj, OBJECT_BODY(obj));
+    if (index < 0)
+        index += elems;
+    return index >= 0 && index < elems && !MVM_is_null(tc, MVM_repr_at_pos_o(tc, obj, index));
 }
 
 MVMint64 MVM_repr_at_pos_i(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx) {
@@ -86,6 +109,98 @@ MVMObject * MVM_repr_at_pos_o(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx
     return NULL;
 }
 
+static void at_pos_multidim(MVMThreadContext *tc, MVMObject *obj, MVMObject *indices, MVMRegister *value, MVMuint16 kind) {
+    MVMint64 num_indices, *c_indices;
+    int_array_to_c_array(tc, indices, &num_indices, &c_indices);
+    REPR(obj)->pos_funcs.at_pos_multidim(tc, STABLE(obj), obj,
+        OBJECT_BODY(obj), num_indices, c_indices, value, kind);
+    MVM_free(c_indices);
+}
+
+MVMint64 MVM_repr_at_pos_multidim_i(MVMThreadContext *tc, MVMObject *obj, MVMObject *indices) {
+    MVMRegister r;
+    at_pos_multidim(tc, obj, indices, &r, MVM_reg_int64);
+    return r.i64;
+}
+
+MVMnum64 MVM_repr_at_pos_multidim_n(MVMThreadContext *tc, MVMObject *obj, MVMObject *indices) {
+    MVMRegister r;
+    at_pos_multidim(tc, obj, indices, &r, MVM_reg_num64);
+    return r.n64;
+}
+
+MVMString * MVM_repr_at_pos_multidim_s(MVMThreadContext *tc, MVMObject *obj, MVMObject *indices) {
+    MVMRegister r;
+    at_pos_multidim(tc, obj, indices, &r, MVM_reg_str);
+    return r.s;
+}
+
+MVMObject * MVM_repr_at_pos_multidim_o(MVMThreadContext *tc, MVMObject *obj, MVMObject *indices) {
+    MVMRegister r;
+    at_pos_multidim(tc, obj, indices, &r, MVM_reg_obj);
+    return r.o;
+}
+
+static void at_pos_2d(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx1, MVMint64 idx2, MVMRegister *value, MVMuint16 kind) {
+    MVMint64 c_indices[2] = { idx1, idx2 };
+    REPR(obj)->pos_funcs.at_pos_multidim(tc, STABLE(obj), obj,
+        OBJECT_BODY(obj), 2, c_indices, value, kind);
+}
+
+MVMint64 MVM_repr_at_pos_2d_i(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx1, MVMint64 idx2) {
+    MVMRegister r;
+    at_pos_2d(tc, obj, idx1, idx2, &r, MVM_reg_int64);
+    return r.i64;
+}
+
+MVMnum64 MVM_repr_at_pos_2d_n(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx1, MVMint64 idx2) {
+    MVMRegister r;
+    at_pos_2d(tc, obj, idx1, idx2, &r, MVM_reg_num64);
+    return r.n64;
+}
+
+MVMString * MVM_repr_at_pos_2d_s(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx1, MVMint64 idx2) {
+    MVMRegister r;
+    at_pos_2d(tc, obj, idx1, idx2, &r, MVM_reg_str);
+    return r.s;
+}
+
+MVMObject * MVM_repr_at_pos_2d_o(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx1, MVMint64 idx2) {
+    MVMRegister r;
+    at_pos_2d(tc, obj, idx1, idx2, &r, MVM_reg_obj);
+    return r.o;
+}
+
+static void at_pos_3d(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx1, MVMint64 idx2, MVMint64 idx3, MVMRegister *value, MVMuint16 kind) {
+    MVMint64 c_indices[3] = { idx1, idx2, idx3 };
+    REPR(obj)->pos_funcs.at_pos_multidim(tc, STABLE(obj), obj,
+        OBJECT_BODY(obj), 3, c_indices, value, kind);
+}
+
+MVMint64 MVM_repr_at_pos_3d_i(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx1, MVMint64 idx2, MVMint64 idx3) {
+    MVMRegister r;
+    at_pos_3d(tc, obj, idx1, idx2, idx3, &r, MVM_reg_int64);
+    return r.i64;
+}
+
+MVMnum64 MVM_repr_at_pos_3d_n(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx1, MVMint64 idx2, MVMint64 idx3) {
+    MVMRegister r;
+    at_pos_3d(tc, obj, idx1, idx2, idx3, &r, MVM_reg_num64);
+    return r.n64;
+}
+
+MVMString * MVM_repr_at_pos_3d_s(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx1, MVMint64 idx2, MVMint64 idx3) {
+    MVMRegister r;
+    at_pos_3d(tc, obj, idx1, idx2, idx3, &r, MVM_reg_str);
+    return r.s;
+}
+
+MVMObject * MVM_repr_at_pos_3d_o(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx1, MVMint64 idx2, MVMint64 idx3) {
+    MVMRegister r;
+    at_pos_3d(tc, obj, idx1, idx2, idx3, &r, MVM_reg_obj);
+    return r.o;
+}
+
 void MVM_repr_bind_pos_i(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx, MVMint64 value) {
     MVMRegister val;
     val.i64 = value;
@@ -114,6 +229,89 @@ void MVM_repr_bind_pos_o(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx, MVM
         idx, val, MVM_reg_obj);
 }
 
+static void bind_pos_multidim(MVMThreadContext *tc, MVMObject *obj, MVMObject *indices, MVMRegister value, MVMuint16 kind) {
+    MVMint64 num_indices, *c_indices;
+    int_array_to_c_array(tc, indices, &num_indices, &c_indices);
+    REPR(obj)->pos_funcs.bind_pos_multidim(tc, STABLE(obj), obj,
+        OBJECT_BODY(obj), num_indices, c_indices, value, kind);
+    MVM_free(c_indices);
+}
+
+void MVM_repr_bind_pos_multidim_i(MVMThreadContext *tc, MVMObject *obj, MVMObject *indices, MVMint64 value) {
+    MVMRegister r;
+    r.i64 = value;
+    bind_pos_multidim(tc, obj, indices, r, MVM_reg_int64);
+}
+void MVM_repr_bind_pos_multidim_n(MVMThreadContext *tc, MVMObject *obj, MVMObject *indices, MVMnum64 value) {
+    MVMRegister r;
+    r.n64 = value;
+    bind_pos_multidim(tc, obj, indices, r, MVM_reg_num64);
+}
+void MVM_repr_bind_pos_multidim_s(MVMThreadContext *tc, MVMObject *obj, MVMObject *indices, MVMString *value) {
+    MVMRegister r;
+    r.s = value;
+    bind_pos_multidim(tc, obj, indices, r, MVM_reg_str);
+}
+void MVM_repr_bind_pos_multidim_o(MVMThreadContext *tc, MVMObject *obj, MVMObject *indices, MVMObject *value) {
+    MVMRegister r;
+    r.o = value;
+    bind_pos_multidim(tc, obj, indices, r, MVM_reg_obj);
+}
+
+static void bind_pos_2d(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx1, MVMint64 idx2, MVMRegister value, MVMuint16 kind) {
+    MVMint64 c_indices[2] = { idx1, idx2 };
+    REPR(obj)->pos_funcs.bind_pos_multidim(tc, STABLE(obj), obj,
+        OBJECT_BODY(obj), 2, c_indices, value, kind);
+}
+
+void MVM_repr_bind_pos_2d_i(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx1, MVMint64 idx2, MVMint64 value) {
+    MVMRegister r;
+    r.i64 = value;
+    bind_pos_2d(tc, obj, idx1, idx2, r, MVM_reg_int64);
+}
+void MVM_repr_bind_pos_2d_n(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx1, MVMint64 idx2, MVMnum64 value) {
+    MVMRegister r;
+    r.n64 = value;
+    bind_pos_2d(tc, obj, idx1, idx2, r, MVM_reg_num64);
+}
+void MVM_repr_bind_pos_2d_s(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx1, MVMint64 idx2, MVMString *value) {
+    MVMRegister r;
+    r.s = value;
+    bind_pos_2d(tc, obj, idx1, idx2, r, MVM_reg_str);
+}
+void MVM_repr_bind_pos_2d_o(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx1, MVMint64 idx2, MVMObject *value) {
+    MVMRegister r;
+    r.o = value;
+    bind_pos_2d(tc, obj, idx1, idx2, r, MVM_reg_obj);
+}
+
+static void bind_pos_3d(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx1, MVMint64 idx2, MVMint64 idx3, MVMRegister value, MVMuint16 kind) {
+    MVMint64 c_indices[3] = { idx1, idx2, idx3 };
+    REPR(obj)->pos_funcs.bind_pos_multidim(tc, STABLE(obj), obj,
+        OBJECT_BODY(obj), 3, c_indices, value, kind);
+}
+
+void MVM_repr_bind_pos_3d_i(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx1, MVMint64 idx2, MVMint64 idx3, MVMint64 value) {
+    MVMRegister r;
+    r.i64 = value;
+    bind_pos_3d(tc, obj, idx1, idx2, idx3, r, MVM_reg_int64);
+}
+void MVM_repr_bind_pos_3d_n(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx1, MVMint64 idx2, MVMint64 idx3, MVMnum64 value) {
+    MVMRegister r;
+    r.n64 = value;
+    bind_pos_3d(tc, obj, idx1, idx2, idx3, r, MVM_reg_num64);
+}
+void MVM_repr_bind_pos_3d_s(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx1, MVMint64 idx2, MVMint64 idx3, MVMString *value) {
+    MVMRegister r;
+    r.s = value;
+    bind_pos_3d(tc, obj, idx1, idx2, idx3, r, MVM_reg_str);
+}
+void MVM_repr_bind_pos_3d_o(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx1, MVMint64 idx2, MVMint64 idx3, MVMObject *value) {
+    MVMRegister r;
+    r.o = value;
+    bind_pos_3d(tc, obj, idx1, idx2, idx3, r, MVM_reg_obj);
+}
+
 void MVM_repr_push_i(MVMThreadContext *tc, MVMObject *obj, MVMint64 pushee) {
     MVMRegister value;
     value.i64 = pushee;
@@ -299,6 +497,35 @@ MVMuint64 MVM_repr_elems(MVMThreadContext *tc, MVMObject *obj) {
     return REPR(obj)->elems(tc, STABLE(obj), obj, OBJECT_BODY(obj));
 }
 
+MVMObject * MVM_repr_dimensions(MVMThreadContext *tc, MVMObject *obj) {
+    if (IS_CONCRETE(obj)) {
+        MVMint64 num_dims, i;
+        MVMint64 *dims;
+        MVMObject *result = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTIntArray);
+        REPR(obj)->pos_funcs.dimensions(tc, STABLE(obj), obj, OBJECT_BODY(obj),
+            &num_dims, &dims);
+        for (i = 0; i < num_dims; i++)
+            MVM_repr_bind_pos_i(tc, result, i, dims[i]);
+        return result;
+    }
+    else {
+        MVM_exception_throw_adhoc(tc, "Cannot get dimensions of a type object");
+    }
+}
+
+MVMint64 MVM_repr_num_dimensions(MVMThreadContext *tc, MVMObject *obj) {
+    if (IS_CONCRETE(obj)) {
+        MVMint64 num_dims;
+        MVMint64 *_;
+        REPR(obj)->pos_funcs.dimensions(tc, STABLE(obj), obj, OBJECT_BODY(obj),
+            &num_dims, &_);
+        return num_dims;
+    }
+    else {
+        MVM_exception_throw_adhoc(tc, "Cannot get number of dimensions of a type object");
+    }
+}
+
 MVMint64 MVM_repr_get_int(MVMThreadContext *tc, MVMObject *obj) {
     if (!IS_CONCRETE(obj))
         MVM_exception_throw_adhoc(tc, "Cannot unbox a type object");
