@@ -186,6 +186,22 @@ static void fix_wval(MVMThreadContext *tc, MVMSpeshGraph *inliner,
     }
 }
 
+/* Resizes the handlers table, making a copy if needed. */
+static void resize_handlers_table(MVMThreadContext *tc, MVMSpeshGraph *inliner, MVMuint32 new_handler_count) {
+    if (inliner->handlers == inliner->sf->body.handlers) {
+        /* Original handlers table; need a copy. */
+        MVMFrameHandler *new_handlers = MVM_malloc(new_handler_count * sizeof(MVMFrameHandler));
+        memcpy(new_handlers, inliner->handlers,
+            inliner->num_handlers * sizeof(MVMFrameHandler));
+        inliner->handlers = new_handlers;
+    }
+    else {
+        /* Probably already did some inlines into this frame; resize. */
+        inliner->handlers = MVM_realloc(inliner->handlers,
+            new_handler_count * sizeof(MVMFrameHandler));
+    }
+}
+
 /* Merges the inlinee's spesh graph into the inliner. */
 static void merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                  MVMSpeshGraph *inlinee, MVMCode *inlinee_code,
@@ -193,6 +209,8 @@ static void merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
     MVMSpeshFacts **merged_facts;
     MVMuint16      *merged_fact_counts;
     MVMint32        i, total_inlines, orig_deopt_addrs;
+    MVMSpeshBB     *inlinee_first_bb = NULL, *inlinee_last_bb = NULL;
+    MVMint32        active_handlers_at_invoke = 0;
 
     /* If the inliner and inlinee are from different compilation units, we
      * potentially have to fix up extra things. */
@@ -267,6 +285,8 @@ static void merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
         }
         bb->idx += inliner->num_bbs - 1; /* -1 as we won't include entry */
         bb->inlined = 1;
+        if (!bb->linear_next)
+            inlinee_last_bb = bb;
         bb = bb->linear_next;
     }
 
@@ -276,7 +296,7 @@ static void merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
     while (bb) {
         if (!bb->linear_next) {
             /* Found the end; insert and we're done. */
-            bb->linear_next = inlinee->entry->linear_next;
+            bb->linear_next = inlinee_first_bb = inlinee->entry->linear_next;
             bb = NULL;
         }
         else {
@@ -402,21 +422,10 @@ static void merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
         inlinee->lexical_types ? inlinee->lexical_types : inlinee->sf->body.lexical_types,
         inlinee->num_lexicals * sizeof(MVMuint16));
 
-    /* Merge handlers. */
+    /* Merge handlers from inlinee. */
     if (inlinee->num_handlers) {
         MVMuint32 total_handlers = inliner->num_handlers + inlinee->num_handlers;
-        if (inliner->handlers == inliner->sf->body.handlers) {
-            /* Original handlers table; need a copy. */
-            MVMFrameHandler *new_handlers = MVM_malloc(total_handlers * sizeof(MVMFrameHandler));
-            memcpy(new_handlers, inliner->handlers,
-                inliner->num_handlers * sizeof(MVMFrameHandler));
-            inliner->handlers = new_handlers;
-        }
-        else {
-            /* Probably already did some inlines into this frame; resize. */
-            inliner->handlers = MVM_realloc(inliner->handlers,
-                total_handlers * sizeof(MVMFrameHandler));
-        }
+        resize_handlers_table(tc, inliner, total_handlers);
         memcpy(inliner->handlers + inliner->num_handlers, inlinee->handlers,
             inlinee->num_handlers * sizeof(MVMFrameHandler));
         for (i = inliner->num_handlers; i < total_handlers; i++) {
@@ -425,12 +434,101 @@ static void merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
         }
     }
 
+    /* If the inliner has handlers in effect at the point of the call that we
+     * are inlining, then we duplicate those and place them surrounding the
+     * inlinee, but with the goto still pointing to the original location.
+     * This means that we can still do a linear scan when searching for an
+     * exception handler, and don't have to try the (costly and fiddly) matter
+     * of trying to traverse the post-inlined call chain. */
+    if (inliner->sf->body.num_handlers) {
+        /* Walk inliner looking for handlers in effect at the point we hit the
+         * invoke instruction we're currently inlining; also record all of the
+         * instructions where the handler "goto" annotation lives. */
+        MVMuint32 orig_handlers = inliner->sf->body.num_handlers;
+        MVMuint8 *active = MVM_spesh_alloc(tc, inliner, orig_handlers);
+        MVMSpeshIns **handler_goto_ins = MVM_spesh_alloc(tc, inliner,
+            orig_handlers * sizeof(MVMSpeshIns *));
+        MVMint32 found_invoke = 0;
+        bb = inliner->entry;
+        while (bb && !bb->inlined) {
+            MVMSpeshIns *ins = bb->first_ins;
+            while (ins) {
+                MVMSpeshAnn *ann = ins->annotations;
+                while (ann) {
+                    if (ann->type == MVM_SPESH_ANN_FH_GOTO) {
+                        if (ann->data.frame_handler_index < orig_handlers)
+                            handler_goto_ins[ann->data.frame_handler_index] = ins;
+                    }
+                    else if (!found_invoke) {
+                        /* Only update these to the point we found the invoke
+                         * being inlined, so it serves as a snapshot of what
+                         * is active. */
+                        if (ann->type == MVM_SPESH_ANN_FH_START)
+                            active[ann->data.frame_handler_index] = 1;
+                        else if (ann->type == MVM_SPESH_ANN_FH_END)
+                            active[ann->data.frame_handler_index] = 0;
+                    }
+                    ann = ann->next;
+                }
+                if (ins == invoke_ins) {
+                    /* Found it; see if we have any handlers active. If so, we
+                     * will continue walking to collect goto annotations. */
+                    found_invoke = 1;
+                    for (i = 0; i < orig_handlers; i++)
+                        active_handlers_at_invoke += active[i];
+                    if (!active_handlers_at_invoke)
+                        break;
+                }
+                ins = ins->next;
+            }
+            if (found_invoke && !active_handlers_at_invoke)
+                break;
+            bb = bb->linear_next;
+        }
+
+        /* If we found handlers active at the point of invoke, duplicate them
+         * in the handlers table and add annotations. */
+        if (active_handlers_at_invoke) {
+            MVMuint32 insert_pos = inliner->num_handlers + inlinee->num_handlers;
+            resize_handlers_table(tc, inliner, insert_pos + active_handlers_at_invoke);
+            for (i = orig_handlers - 1; i >= 0; i--) {
+                if (active[i]) {
+                    /* Add handler start annotation to first inlinee instruction. */
+                    MVMSpeshAnn *new_ann = MVM_spesh_alloc(tc, inliner, sizeof(MVMSpeshAnn));
+                    new_ann->type = MVM_SPESH_ANN_FH_START;
+                    new_ann->data.frame_handler_index = insert_pos;
+                    new_ann->next = inlinee_first_bb->first_ins->annotations;
+                    inlinee_first_bb->first_ins->annotations = new_ann;
+
+                    /* Add handler end annotation to last inlinee instruction. */
+                    new_ann = MVM_spesh_alloc(tc, inliner, sizeof(MVMSpeshAnn));
+                    new_ann->type = MVM_SPESH_ANN_FH_END;
+                    new_ann->data.frame_handler_index = insert_pos;
+                    new_ann->next = inlinee_last_bb->last_ins->annotations;
+                    inlinee_last_bb->last_ins->annotations = new_ann;
+
+                    /* Add handler goto annotation to original target in inliner. */
+                    new_ann = MVM_spesh_alloc(tc, inliner, sizeof(MVMSpeshAnn));
+                    new_ann->type = MVM_SPESH_ANN_FH_GOTO;
+                    new_ann->data.frame_handler_index = insert_pos;
+                    new_ann->next = handler_goto_ins[i]->annotations;
+                    handler_goto_ins[i]->annotations = new_ann;
+
+                    /* Copy handler entry to new slot. */
+                    memcpy(inliner->handlers + insert_pos, inliner->handlers + i,
+                        sizeof(MVMFrameHandler));
+                    insert_pos++;
+                }
+            }
+        }
+    }
+
     /* Update total locals, lexicals, basic blocks, and handlers of the
      * inliner. */
     inliner->num_bbs      += inlinee->num_bbs - 1;
     inliner->num_locals   += inlinee->num_locals;
     inliner->num_lexicals += inlinee->num_lexicals;
-    inliner->num_handlers += inlinee->num_handlers;
+    inliner->num_handlers += inlinee->num_handlers + active_handlers_at_invoke;
 }
 
 /* Tweak the successor of a BB, also updating the target BBs pred. */
