@@ -41,6 +41,7 @@ typedef struct MVMExtOpRegistry MVMExtOpRegistry;
 typedef struct MVMExtRegistry MVMExtRegistry;
 typedef struct MVMFixedSizeAlloc MVMFixedSizeAlloc;
 typedef struct MVMFixedSizeAllocFreeListEntry MVMFixedSizeAllocFreeListEntry;
+typedef struct MVMFixedSizeAllocSafepointFreeListEntry MVMFixedSizeAllocSafepointFreeListEntry;
 typedef struct MVMFixedSizeAllocSizeClass MVMFixedSizeAllocSizeClass;
 typedef struct MVMFrame MVMFrame;
 typedef struct MVMFrameHandler MVMFrameHandler;
@@ -100,6 +101,9 @@ typedef struct MVMCUnionREPRData MVMCUnionREPRData;
 typedef struct MVMMultiCache MVMMultiCache;
 typedef struct MVMMultiCacheBody MVMMultiCacheBody;
 typedef struct MVMMultiArityCache MVMMultiArityCache;
+typedef struct MVMMultiDimArray MVMMultiDimArray;
+typedef struct MVMMultiDimArrayBody MVMMultiDimArrayBody;
+typedef struct MVMMultiDimArrayREPRData MVMMultiDimArrayREPRData;
 typedef struct MVMContinuation MVMContinuation;
 typedef struct MVMContinuationBody MVMContinuationBody;
 typedef struct MVMReentrantMutex MVMReentrantMutex;
@@ -186,7 +190,7 @@ typedef struct MVMIOAsyncReadable MVMIOAsyncReadable;
 typedef struct MVMIOAsyncWritable MVMIOAsyncWritable;
 typedef struct MVMIOSeekable MVMIOSeekable;
 typedef struct MVMIOSockety MVMIOSockety;
-typedef struct MVMIOInteractive MVMIOInteractive;
+typedef struct MVMIOPipeable MVMIOPipeable;
 typedef struct MVMIOLockable MVMIOLockable;
 typedef struct MVMIOSyncStreamData MVMIOSyncStreamData;
 typedef struct MVMIOSyncPipeData MVMIOSyncPipeData;
